"""Implements twitter."""

import argparse
import logging
import tweepy

from utils import timex

MAX_LEN_TWEET = 280
MAX_MEDIA_FILES = 4

log = logging.getLogger('twitter-wrapper')
logging.basicConfig(level=logging.INFO)


class Twitter:
    """Implements Twitter wrapper."""

    def __init__(
        self,
        twtr_api_key,
        twtr_api_secret_key,
        twtr_access_token,
        twtr_access_token_secret,
    ):
        """Construct Twitter."""
        if twtr_api_key:
            auth = tweepy.OAuthHandler(twtr_api_key, twtr_api_secret_key)
            auth.set_access_token(twtr_access_token, twtr_access_token_secret)
            self.api = tweepy.API(auth)
            api_me = self.api.me()
            log.info(
                'Created Twitter for @%s (%s).',
                api_me.screen_name,
                api_me.id_str
            )
        else:
            log.error(
                'Missing twitter API Key etc. Cannot create Twitter.',
            )
            self.api = None

    @staticmethod
    def from_args(description=''):
        """Construct Twitter from Args."""
        parser = argparse.ArgumentParser(
            description=description,
        )
        for twtr_arg_name in [
            'twtr_api_key',
            'twtr_api_secret_key',
            'twtr_access_token',
            'twtr_access_token_secret',
        ]:
            parser.add_argument(
                '--' + twtr_arg_name,
                type=str,
                required=False,
                default=None,
            )
        args = parser.parse_args()
        return Twitter(
            args.twtr_api_key,
            args.twtr_api_secret_key,
            args.twtr_access_token,
            args.twtr_access_token_secret,
        )

    def tweet(
        self,
        tweet_text,
        status_image_files=None,
        update_user_profile=False,
        profile_image_file=None,
        banner_image_file=None,
    ):
        """Tweet."""
        if status_image_files is None:
            status_image_files = []

        if not self.api:
            log.error('Missing API. Cannot tweet')
            return False

        if len(tweet_text) > MAX_LEN_TWEET:
            log.error('Tweet text is too long. Not tweeting.')
            return

        if len(status_image_files) > MAX_MEDIA_FILES:
            log.warning('Too many (%d) status image files. Truncating.')
            status_image_files = status_image_files[:MAX_MEDIA_FILES]

        media_ids = []
        for image_file in status_image_files:
            res = self.api.media_upload(image_file)
            media_id = res.media_id
            media_ids.append(media_id)
            log.info(
                'Uploaded status image %s to twitter as %s',
                image_file,
                media_id,
            )

        if media_ids:
            self.api.update_status(tweet_text, media_ids=media_ids)
        else:
            self.api.update_status(tweet_text)
        log.info('Updated status to: %s', tweet_text)

        if update_user_profile:
            date = timex.format_time(
                timex.get_unixtime(),
                '%B %d, %Y %H:%M%p',
            )
            timezone = timex.get_timezone()
            description = 'Automatically updated at {date} {timezone}'\
                .format(date=date, timezone=timezone)
            self.api.update_profile(description=description)
            log.info('Updated profile description to: %s', description)

        if profile_image_file:
            self.api.update_profile_image(profile_image_file)
            log.info(
                'Update profile image to %s',
                profile_image_file,
            )

        if banner_image_file:
            self.api.update_profile_banner(banner_image_file)
            log.info(
                'Update profile banner image to %s',
                banner_image_file,
            )
        return True


if __name__ == '__main__':
    twtr = Twitter.from_args()
    twtr.tweet('Test')



"""Filex."""


def read(file_name):
    """Read."""
    with open(file_name, 'r') as fin:
        content = fin.read()
        fin.close()
        return content


def write(file_name, content, mode='w'):
    """Write."""
    with open(file_name, mode) as fout:
        fout.write(content)
        fout.close()



"""Utils related to lists and dicts."""


def dict_list_to_index(dict_list, key):
    """Given a list of dicts, returns an index mapping each dict item
    to the dict.

    Args:
        dict_list (list of dicts): list of dicts
        key (str): key

    Return:
        index (dict)

    .. code-block:: python

        >>> from utils.ds import dict_list_to_index
        >>> dict_list = [
            {'name': 'Alice', 'age': 20},
            {'name': 'Bob', 'age': 25},
            {'name': 'Charlie', 'age': 55},
        ]
        >>> print(dict_list_to_index(dict_list, 'name'))
        {
            'Alice': {'name': 'Alice', 'age': 20},
            'Bob': {'name': 'Bob', 'age': 25},
            'Charlie': {'name': 'Charlie', 'age': 55},
        }


    """
    return dict(zip(
        list(map(
            lambda d: d.get(key, None),
            dict_list,
        )),
        dict_list,
    ))


def unique(lst):
    """Get unique values from list.

    Args:
        lst (list): List

    Return:
        list of unique values

    .. code-block:: python

        >>> from utils.ds import unique
        >>> lst = [1, 1, 1, 2]
        >>> print(unique(lst))
        [1, 2]

    """
    return list(set(lst))


def flatten(list_of_list):
    """Flatten list of lists.

    Args:
        list_of_list(list): list of lists

    Return:
        flattened list

    .. code-block:: python

    >>> print(flatten([[1, 2], [3, 4, 5], [6], []]))
    [1, 2, 3, 4, 5, 6]

    """
    flattened_list = []
    for lst in list_of_list:
        flattened_list += lst
    return flattened_list


def sort_dict_items_by_key(_dict):
    """Sort dict items by key."""
    return sorted(
        _dict.items(),
        key=lambda item: item[0],
    )



"""Setup."""
import time
import setuptools

DIST_NAME = 'utils'
long_description = ''
with open('README.md') as fin:
    long_description = fin.read()

IS_PRE_RELEASE = False
MAJOR, MINOR, PATCH = 1, 0, 14

if IS_PRE_RELEASE:
    ts = time.strftime('%Y%m%d%H%M%S0000', time.localtime())
    version = '%d.%d.%d.%s' % (MAJOR, MINOR, PATCH, ts)
else:
    version = '%d.%d.%d' % (MAJOR, MINOR, PATCH)

setuptools.setup(
    name="%s-nuuuwan" % DIST_NAME,
    version=version,
    author="Nuwan I. Senaratna",
    author_email="nuuuwan@gmail.com",
    description="Simple extensions to the core python libraries.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nuuuwan/%s" % DIST_NAME,
    project_urls={
        "Bug Tracker": "https://github.com/nuuuwan/%s/issues" % DIST_NAME,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        'area',
        'bs4',
        'psutil',
        'requests',
        'selenium',
        'tweepy',
        'pandas',
        'shapely',
        'geopandas',
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
)



"""Utils for caching functions."""
import hashlib
import json
import os
import threading
import time
import base64

from functools import wraps

from geopandas.geodataframe import GeoDataFrame
from pandas.core.frame import DataFrame
from shapely.geometry.base import BaseGeometry
from shapely.geometry import shape, mapping

from utils import filex

CACHE_DIR = '/tmp/cache'
CACHE_DEFAULT_TIMEOUT = 60


def _json_serialize(data):
    if isinstance(data, bytes):
        return {
            'type': 'bytes',
            'data': base64.b64encode(data).decode('ascii'),
        }
    if isinstance(data, GeoDataFrame):
        return {
            'type': 'GeoDataFrame',
            'data': data.to_dict(),
        }
    if isinstance(data, DataFrame):
        return {
            'type': 'DataFrame',
            'data': data.to_dict(),
        }
    if isinstance(data, BaseGeometry):
        return {
            'type': 'BaseGeometry',
            'data': mapping(data),
        }

    return {
        'type': None,
        'data': data,
    }


def _json_deserialize(data):
    data_type = data['type']
    data_data = data['data']

    if data_type == 'bytes':
        return base64.b64decode(data_data)
    if data_type == 'DataFrame':
        return DataFrame.from_dict(data_data)
    if data_type == 'GeoDataFrame':
        return GeoDataFrame.from_dict(data_data)
    if data_type == 'BaseGeometry':
        return shape(data_data)

    return data_data


class _Cache:
    """Implements base cache logic.

    Data is cached both in memory and, for persistance, on disk. On *get*
    requests, memory is first inspected. If cache key is not found there,
    disk is inspected.

    Args:
        cache_name(str): Name to identify cache.
        timeout (int): timeout in seconds

    Note:
        Used by decorator *cache* (See below).

    """

    @staticmethod
    def __get_hash(cache_key):
        return hashlib.md5(cache_key.encode()).hexdigest()

    __dir = '/tmp/cache'
    __store = {}
    __lock_map_lock = threading.Lock()
    __lock_map = {}
    __cache_name = 'new_cache'

    def __init__(self, cache_name, timeout=CACHE_DEFAULT_TIMEOUT):
        """Implement class constructor."""
        self.__cache_name = cache_name
        self.__timeout = timeout
        os.system('mkdir -p %s' % self.__get_dir())

    def __get_dir(self):
        return '%s/%s' % (CACHE_DIR, self.__cache_name)

    def __get_lock(self, key):
        # pylint: disable=R1732
        self.__lock_map_lock.acquire()
        if key not in self.__lock_map:
            self.__lock_map[key] = threading.Lock()
        lock = self.__lock_map[key]
        self.__lock_map_lock.release()
        return lock

    def __acquire_lock(self, key):
        self.__get_lock(key).acquire()

    def __release_lock(self, key):
        self.__get_lock(key).release()

    def __get_cache_file_name(self, cache_key):
        return '%s/%s' % (
            self.__get_dir(),
            _Cache.__get_hash(cache_key),
        )

    def __get_file_exists(self, key):
        return os.path.exists(self.__get_cache_file_name(key))

    def __get_from_file(self, key):
        data_json = filex.read(self.__get_cache_file_name(key))

        if data_json == '':
            return None
        packet = json.loads(data_json)
        return packet

    def __set(self, key, data):
        self.__acquire_lock(key)
        packet = {
            'data': _json_serialize(data),
            'set_time': time.time(),
        }
        filex.write(
            self.__get_cache_file_name(key),
            json.dumps(packet, ensure_ascii=True),
        )
        self.__store[key] = packet
        self.__release_lock(key)

    def get(self, key_or_list, fallback):
        """Get data from cache if cache key exists, if not from fallback.

        Args:
            key_or_list (str or list): the cache key can be given as a string,
                or a list of strings, which are joined
            fallback (function): function to call of cache key not in cache

        Returns:
            data

        """
        if isinstance(key_or_list, list):
            key = ':'.join(key_or_list)
        else:
            key = key_or_list

        packet = None

        if key in self.__store:
            packet = self.__store[key]

        if not packet:
            if self.__get_file_exists(key):
                packet = self.__get_from_file(key)
                if packet is not None:
                    self.__store[key] = packet

        if packet:
            if 'set_time' in packet:
                min_set_time = time.time() - self.__timeout
                if packet['set_time'] > min_set_time:
                    return _json_deserialize(packet['data'])

        data = fallback()
        if data is not None:
            self.__set(key, data)
        return data

    def flush_store(self):
        """Flush all cache."""
        self.__store = {}

    def flush(self, key):
        """Flush cache for a given cache key.

        Args:
            key (str): Cache key to flush
        """
        self.__acquire_lock(key)
        if self.__get_file_exists(key):
            os.remove(self.__get_cache_file_name(key))
        if key in self.__store:
            del self.__store[key]
        self.__release_lock(key)


def cache(cache_name, timeout=CACHE_DEFAULT_TIMEOUT):
    """Wrap class Cache as decorator.

    Args:
        cache_name (str): cache name
        timeout (int): timeout in seconds

    .. code-block:: python

        from utils.cache import cache

        @cache('test', 86400)
        def long_operation():
            import time
            time.sleep(100)
            return 1

        >>> long_operation() # takes >100s to run
        >>> long_operation() # runs almost instantly

    """

    def cache_inner(func):

        @wraps(func)
        def cache_inner_inner(*args, **kwargs):
            cache_key = json.dumps({
                'cache_name': cache_name,
                'function_name': func.__name__,
                'kwargs': kwargs,
                'args': args,
            })

            def fallback():
                return func(*args, **kwargs)

            return _Cache(cache_name, timeout).get(cache_key, fallback)

        return cache_inner_inner
    return cache_inner
